/* stablesort.c -- an adaptive, stable, natural mergesort

   http://py-stablesort.sourceforge.net/

   Wrapper code Copyright (C) 2002-2003 Markus F.X.J. Oberhumer
   <mfx@users.sourceforge.net>

   Actual sort implementation Copyright (C) 2002-2003 the
   Python Software Foundation. Written by Tim Peters.
 */


#define MODULE_VERSION  "1.2.2"

#include <Python.h>
#undef NDEBUG
#define NDEBUG
#include <assert.h>

#if !defined(PY_VERSION_HEX) || (PY_VERSION_HEX < 0x010502f0)
#  error "Need Python version 1.5.2 or better"
#endif

#undef UNUSED
#define UNUSED(var)     ((void)&var)


/***********************************************************************
// compat stuff
************************************************************************/

#if (PY_VERSION_HEX < 0x01060000)
#  define PyMem_FREE(a)         PyMem_DEL(a)
#endif

#if (PY_VERSION_HEX < 0x02020000)
#  define PyObject_Call(a,b,c)  PyEval_CallObjectWithKeywords(a,b,c)
#endif

#if (PY_VERSION_HEX < 0x02010000)
static int Compare_LT(PyObject *a, PyObject *b)
{
    int x;
    x = PyObject_Compare(a, b);
    if (PyErr_Occurred())
        return -1;
    return x < 0;
}
#endif


/***********************************************************************
// wrapper
************************************************************************/

#define STABLESORT
#include "listobject.c"
#undef STABLESORT
#undef list


static PyMethodDef *list_sort_md = NULL;
static PyMethodDef saved_list_sort;


#define SORT_DOC \
"sort(list, cmpfunc=None) -- stable sort *IN PLACE*; cmpfunc(x, y) -> -1, 0, 1"
#define INSTALL_DOC \
"install() -- install the stablesort implementation as global list.sort() method"
#define UNINSTALL_DOC \
"uninstall() -- restore the original list.sort() implementation"

/* note: we use "X.sort" instead of "L.sort" so that you can see
 * that the new method is installed */
#define SORT_METH_DOC \
"X.sort(cmpfunc=None) -- stable sort *IN PLACE*; cmpfunc(x, y) -> -1, 0, 1"


/***********************************************************************
// stablesort.sort(list, cmpfunc=None)  [function]
************************************************************************/

static PyObject *
do_sort(PyObject *dummy, PyObject *args)
{
    PyListObject *list = NULL;
    PyObject *cmpfunc = NULL;

    UNUSED(dummy);
    if (!PyArg_ParseTuple(args, "O!|O:sort", &PyList_Type, &list, &cmpfunc))
        return NULL;
    return listsort(list, cmpfunc);
}


/***********************************************************************
// list.sort(cmpfunc=None)          [method]
// stablesort.install()             [function]
// stablesort.uninstall()           [function]
************************************************************************/

static PyObject *
do_new_list_sort(PyListObject *self, PyObject *args)
{
    PyObject *cmpfunc = NULL;

    if (!PyArg_ParseTuple(args, "|O:sort", &cmpfunc))
        return NULL;
    return listsort(self, cmpfunc);
}


static PyObject *
do_install(PyObject *dummy, PyObject *args)
{
    UNUSED(dummy);
    if (!PyArg_ParseTuple(args, ":install"))
        return NULL;
    if (list_sort_md->ml_meth != (PyCFunction)do_new_list_sort)
    {
        memcpy(&saved_list_sort, list_sort_md, sizeof(PyMethodDef));
        list_sort_md->ml_meth = (PyCFunction)do_new_list_sort;
        list_sort_md->ml_flags = METH_VARARGS;
        list_sort_md->ml_doc = SORT_METH_DOC;
    }
    Py_INCREF(Py_None);
    return Py_None;
}


static PyObject *
do_uninstall(PyObject *dummy, PyObject *args)
{
    UNUSED(dummy);
    if (!PyArg_ParseTuple(args, ":uninstall"))
        return NULL;
    if (list_sort_md->ml_meth == (PyCFunction)do_new_list_sort)
        memcpy(list_sort_md, &saved_list_sort, sizeof(PyMethodDef));
    Py_INCREF(Py_None);
    return Py_None;
}


/***********************************************************************
// Small hack to get access to Python's internal list.sort() method.
// Note that the old immutable_list_type has gone since Python 2.3a1.
************************************************************************/

static int init_internals(void)
{
    int result = -1;        /* guilty until proved innocent */
    PyObject *list = NULL;
    PyObject *sort_meth = NULL;
    PyCFunctionObject *m;

    /* Create a list. */
    list = PyList_New(0);
    if (list == NULL)
        goto error;

    /* Get the original list.sort() method. */
    sort_meth = PyObject_GetAttrString(list, "sort");
    if (sort_meth == NULL || !PyCFunction_Check(sort_meth))
        goto error;

    m = (PyCFunctionObject *) sort_meth;
    if (m->m_ml && m->m_ml->ml_meth)
    {
        /* Permanently leak a reference so that the original sort_meth will
         * never get garbage collected. */
        sort_meth = NULL;
        /* success */
        list_sort_md = m->m_ml;
        result = 0;
    }

error:
    Py_XDECREF(sort_meth);
    Py_XDECREF(list);
    return result;
}


/***********************************************************************
// main
************************************************************************/

static PyMethodDef methods[] =
{
    {"sort",       (PyCFunction)do_sort,      METH_VARARGS, SORT_DOC},
    {"install",    (PyCFunction)do_install,   METH_VARARGS, INSTALL_DOC},
    {"uninstall",  (PyCFunction)do_uninstall, METH_VARARGS, UNINSTALL_DOC},
    {NULL, NULL, 0, NULL}
};


static /* const */ char module_documentation[] =
    SORT_DOC "\n"
    INSTALL_DOC "\n"
    UNINSTALL_DOC "\n"
;


#if defined(PyMODINIT_FUNC)
PyMODINIT_FUNC
#else
#if defined(__cplusplus)
extern "C"
#endif
DL_EXPORT(void)
#endif
initstablesort(void)
{
    PyObject *m, *d, *v;

    if (init_internals() != 0) {
        /* this really should not happen... */
        Py_FatalError("stablesort: init_internals() failed !!");
        return;
    }

    m = Py_InitModule4("stablesort", methods, module_documentation,
                       NULL, PYTHON_API_VERSION);
    if (!m) goto error;
    d = PyModule_GetDict(m);
    if (!d) goto error;

    v = PyString_FromString("Markus F.X.J. Oberhumer (wrapper), Tim Peters (actual sort implementation)");
    if (!v) goto error;
    PyDict_SetItemString(d, "__author__", v);
    Py_DECREF(v);
    v = PyString_FromString(MODULE_VERSION " (listobject.c CVS 2.156)");
    if (!v) goto error;
    PyDict_SetItemString(d, "__version__", v);
    Py_DECREF(v);
    v = PyString_FromString(__DATE__ " " __TIME__);
    if (!v) goto error;
    PyDict_SetItemString(d, "__compile_date__", v);
    Py_DECREF(v);
    return;

error:
    Py_FatalError("stablesort: initstablesort() failed !!");
}


/*
vi:ts=4:et
*/
